local menupath = core.get_mainmenu_path( )
local basepath = core.get_builtin_path( )
defaulttexturedir = core.get_texturepath_share( ) .. DIR_DELIM .. "base" ..
					DIR_DELIM .. "pack" .. DIR_DELIM

--------------------------------------------------------------------------------

config = {
	status_url = "http://www.clubequinox.com/status.txt",
	create_url = "http://www.clubequinox.com/api/create.lua",
	launch_url = "http://www.virtualdanceclub.com/launch/index.html",
	policy_url = "http://www.clubequinox.com/policy.html",
	remote_addr = "game.clubequinox.com",
	remote_port = "30010",
}

local http_req = core.get_http_api( )
local fs = minetest.formspec_escape
local sprintf = string.format
local cur_date = os.date( "*t" )
local server_status

local function get_numeric_range( a, b )
	local list = { }
	for n = a, b, b < a and -1 or 1 do
		table.insert( list, n )
	end
	return list
end

local function clamp( val, min, max )
        return val < min and min or val > max and max or val
end

local function http_request( url, headers, fields, timeout )
	return http_req.fetch_sync( { url = url, timeout = timeout, extra_headers = headers, post_data = fields } )
end

local function ServerStatus( )
	local self = { }
	local status

	self.refresh = function ( )
		local results = http_request( config.status_url, nil, nil, 5.0 )

		if results.timeout or results.code ~= 200 then
			status = nil
		else
			status = core.parse_json( results.data )
		end
	end

	self.get_formspec = function ( )
		if status.is_online then
			local formspec =
				"image[14.2,18.5;7.737,1.805;" .. fs( defaulttexturedir .. "menu_status.png" ) .. "]" ..
				"image[15.5,20.2;5.029,1.032;" .. fs( defaulttexturedir .. "menu_online.png" ) .. "]" ..
				sprintf( "hypertext[14.0,21.5;4.0,1.0;;<global size=12><b>Cur Users:</b> %s]", status.cur_users ) ..
				sprintf( "hypertext[18.0,21.5;4.0,1.0;;<global size=12 halign=right><b>Uptime:</b> %d min]", status.uptime / 60 )
			return formspec
		else
			local formspec =
				"image[14.2,18.5;7.737,1.805;" .. fs( defaulttexturedir .. "menu_status.png" ) .. "]" ..
				"image[15.5,20.2;5.029,1.032;" .. fs( defaulttexturedir .. "menu_offline.png" ) .. "]" ..
				"hypertext[14.0,21.5;8.0,1.0;;<global size=12 halign=center><b>Please check again later.</b>]"
			return formspec
		end
	end

	self.refresh( )

	return self
end

local function Birthdate( )
	local self = { day = 1, month = 1, year = cur_date.year }

	local function days_from_civil( y, m, d )
		-- https://stackoverflow.com/questions/28550413/compare-dates-in-lua
		-- hack because Windows cannot support timestamps before 1970
		if m <= 2 then
			y = y - 1
			m = m + 9
		else
			m = m - 3
		end
		local era = math.floor( y / 400 )
		local yoe = y - era * 400                                           -- [0, 399]
		local doy = math.modf( ( 153 * m + 2 ) / 5 ) + d - 1                          -- [0, 365]
		local doe = yoe * 365 + math.modf( yoe / 4 ) - math.modf( yoe / 100 ) + doy -- [0, 146096]
		return era * 146097 + doe - 719468
	end

	self.set_day = function ( day )
		self.day = clamp( day, 1, 31 )
	end

	self.set_month = function ( month )
		self.month = clamp( month, 1, 12 )
	end

	self.set_year = function ( year )
		self.year = clamp( year, cur_date.year - 100, cur_date.year )
	end

	self.to_age = function ( )
		return ( days_from_civil( cur_date.year, cur_date.month, cur_date.day ) - days_from_civil( self.year, self.month, self.day ) ) / 365
	end

	self.to_birthday = function ( )
		return days_from_civil( self.year, self.month, self.day )
	end

	return self
end

local function ContentViewer( width, height, bgcolor, bgimage, margin )
	local self = { }
	local p1, p2

	self.recalc_window = function ( )
		local window_width = core.get_screen_info( ).window_width
		local portal_width = math.min( 1012, window_width ) * 36 / 1012

		p1 = ( 36 - portal_width ) / 2 + margin
		p2 = 36 - ( 36 - portal_width ) / 2 - margin
	end

	self.get_formspec_prefix = function ( )
		local formspec = sprintf( "size[%0.2f,%0.2f;true]", width,  height ) ..
			"position[0.5,0.5]" ..
			"anchor[0.5,0.5]" ..
			"real_coordinates[true]" ..
			sprintf( "bgcolor[%s;true]", bgcolor ) ..
			"background[0.0,-1.0;36.0,24.0;" .. fs( defaulttexturedir .. bgimage ) .. ";true]"

		return formspec
	end

	self.get_formspec_header = function ( step_num, step_title )
		local formspec = 
			sprintf( "image[%0.1f,2.6;2.579,2.579;%s]", p1, fs( defaulttexturedir .. "menu_logo.png" ) ) ..
			sprintf( "hypertext[%0.1f,5.4;6.0,2.0;;<global size=14><b>Club Equinox</b>]", p1 - 0.1 ) ..
			sprintf( "hypertext[%0.1f,6.0;6.0,2.0;;<global size=14>VR Client 1.0]", p1 - 0.1 ) ..
			sprintf( "image[%0.1f,2.5;2.502,2.192;%s]", p2 - 2.5, fs( defaulttexturedir .. "menu_step" .. step_num .. ".png" ) ) ..
			sprintf( "hypertext[%0.1f,4.7;6.0,2.0;;<global size=14 halign=right><b>%s</b>]", p2 - 5.9, step_title )

		return formspec
	end

	self.get_formspec_header2 = function ( step_num, step_title )
		local formspec = 
			sprintf( "image[%0.1f,2.6;2.579,2.579;%s]", p1, fs( defaulttexturedir .. "menu_logo.png" ) ) ..
			sprintf( "hypertext[%0.1f,5.4;6.0,2.0;;<global size=14><b>Club Equinox</b>]", p1 - 0.1 ) ..
			sprintf( "hypertext[%0.1f,6.0;6.0,2.0;;<global size=14>VR Client 1.0]", p1 - 0.1 )

		return formspec
	end

	self.get_formspec_footer = function ( )
		local formspec = 
			"image[2.6,2.5;2.579,2.579;" .. fs( defaulttexturedir .. "menu_logo.png" ) .. "]" ..
			"hypertext[2.5,5.4;6.0,2.0;;<b>ClubEquinox Client</b> Version 1.0]"
		return formspec
	end

	self.recalc_window( )

	return self
end

function open_alert_box( type, message, on_confirm )
	local viewer = ContentViewer( 36, 24, "#880066FF", "background2.png", 1.5 )

	local function get_formspec( )
		local formspec =
			viewer.get_formspec_prefix( ) ..
			viewer.get_formspec_header2( ) ..
			sprintf( "image[12.5,9.0;2.0,2.0;%s]", fs( defaulttexturedir .. "alert_" .. type .. ".png" ) ) ..
			sprintf( "textarea[15.0,9.0;9.0,4.0;;%s;]", fs( message ) ) ..
			"button[16.5,14.5;3.0,1.0;okay;Okay]"
		return formspec
	end

	local function on_close( state, fields )
		if fields.quit then return end  -- ignore signals

		if fields.okay then
			minetest.destroy_menu( )
			if on_confirm then on_confirm( ) end
		end
	end

	minetest.create_menu( nil, get_formspec( ), on_close, minetest.FORMSPEC_SIGHOLD )
end

function open_dialog_box( type, message, on_confirm, options )
	local viewer = ContentViewer( 36, 24, "#880066FF", "background2.png", 1.5 )

	local function get_formspec( )

		local formspec = 
			viewer.get_formspec_prefix( ) ..
			viewer.get_formspec_header2( ) ..
			sprintf( "image[12.5,9.0;2.0,2.0;%s]", fs( defaulttexturedir .. "alert_" .. type .. ".png" ) ) ..
			sprintf( "textarea[15.0,9.0;9.0,4.0;;%s;]", fs( message ) ) ..
			"button[13.9,14.5;3.0,1.0;no;No]" ..
			"button[18.1,14.5;3.0,1.0;yes;Yes]"
		return formspec
	end

	local function on_close( state, fields )
		if fields.quit then return end  -- ignore signals

		if fields.yes or fields.no then
			minetest.destroy_menu( )
			if on_confirm then
				on_confirm( fields.yes and "yes" or "no" )
			end
		end
	end

	minetest.create_menu( nil, get_formspec( ), on_close, minetest.FORMSPEC_SIGHOLD )
end

local function register_account( username, password, birthday )
        local fields = {
                username = username,
                password = password,
                birthday = birthday,
        }

	local results = http_request( config.create_url, nil, fields, 5.0 )

	if results.timeout then
		return "Error: Connection timed out."

	elseif results.code ~= 200 then
		if results.code == 404 then
			return "Error: Resource not found (status 404)."
		elseif results.code == 403 then
			return "Error: Permission denied (status 403)."
		elseif results.code == 500 then
			return "Error: Internal server error (status 500)."
		elseif results.code == 504 then
			return "Error: Gateway timed out (status 504)."
		end
	end

	local status = core.parse_json( results.data )

	if not status then
		return "The request could not be processed at this time."
	elseif status.error then
		return "An account with this username already exists."
	end
end

local function open_connect_viewer( )
	local viewer = ContentViewer( 36, 24, "#880066FF", "background.png", 1.5 )
	local username = ""

	local function get_formspec( )
		local formspec =
			viewer.get_formspec_prefix( ) ..
			viewer.get_formspec_header( 2, "Join the Game" ) ..
			"image[14.2,5.5;7.737,2.192;" .. fs( defaulttexturedir .. "menu_title2.png" ) .. "]" ..
			"hypertext[12.0,8.0;12.0,3.0;menu_text;<tag name=action color=#DD44AA>Enter your account information to launch the audio stream and join the game.]" ..
			"label[12.5,10.2;Username:]" ..
			"label[12.5,11.2;Password:]" ..
			sprintf( "field[16.5,9.8;5.5,0.8;username;;%s]", username ) ..
			"pwdfield[16.5,10.8;5.5,0.8;password;]" ..
			"hypertext[12.0,12.0;12.0,3.0;menu_text;<tag name=action color=#DD44AA>If you need to create a new account, then <action name=backstep>return to step 1</action>.]" ..
			"button[16.5,14.0;3.0,1.0;register;Connect]" ..
			server_status.get_formspec( )


		return formspec
	end

	local function on_close( state, fields )
		if fields.quit == minetest.FORMSPEC_SIGCONT then
			minetest.update_menu( get_formspec( ) )
			return
		elseif fields.quit then
			return  -- ignore other signals
		end

		if fields.menu_text == "action:backstep" then
			open_register_viewer( )

		elseif fields.register then
			username = fields.username

			if #fields.username < 3 then
				open_alert_box( "halt", "Error:\n\nPlease enter your username." )
				return
			elseif #fields.username < 3 then
				open_alert_box( "halt", "Error:\n\nPlease enter your password." )
				return
			end

			core.open_url( config.launch_url .. "?" .. fields.username )

			gamedata.playername = fields.username
			gamedata.password   = fields.password
			gamedata.address    = config.remote_addr
			gamedata.port       = config.remote_port
			gamedata.selected_world = 0

			core.start( )
		end
	end

	server_status.refresh( )

	core.create_menu( nil, get_formspec( ), on_close, core.FORMSPEC_SIGSTOP )	
end

local function open_register_viewer( )
	local viewer = ContentViewer( 36, 24, "#880066FF", "background.png", 1.5 )
	local day_list = get_numeric_range( 1, 31 )
	local month_list = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" }
	local year_list = get_numeric_range( cur_date.year, cur_date.year - 90 )
	local username = ""
	local birthdate = Birthdate( )
	local is_account_created = false

	local function get_formspec( )
		local formspec =
			viewer.get_formspec_prefix( ) ..
			viewer.get_formspec_header( 1, "Register Account" ) ..
			"image[14.2,5.5;7.737,2.192;" .. fs( defaulttexturedir .. "menu_title1.png" ) .. "]" ..
			"hypertext[12.0,8.0;12.0,3.0;menu_text;<tag name=action color=#DD44AA>Before joining the game, you will need to create an account. If you already have an account, then you can <action name=skipstep>skip this step</action>.]" ..
			"label[12.5,11.2;Username:]" ..
			"label[12.5,12.2;Password:]" ..
			"label[12.5,13.2;Birth Date:]" ..
			sprintf( "field[16.5,10.8;5.5,0.8;username;;%s]", username ) ..
			"pwdfield[16.5,11.8;5.5,0.8;password;]" ..
			sprintf( "dropdown[16.5,12.8;1.5,0.8;birthdate_day;%s;%d;true]", table.concat( day_list, "," ), birthdate.day ) ..
			sprintf( "dropdown[18.0,12.8;2.0,0.8;birthdate_month;%s;%d;true]", table.concat( month_list, "," ), birthdate.month ) ..
			sprintf( "dropdown[20.0,12.8;2.5,0.8;birthdate_year;%s;%d;false]", table.concat( year_list, "," ), cur_date.year - birthdate.year + 1 ) ..
			"button[16.5,14.0;3.0,1.0;register;Register]" ..
			"hypertext[12.0,15.4;12.0,3.0;menu_text;<tag name=action color=#DD44AA>By creating an account, you agree to the <action name=policy>Acceptable Use Policy</action> posted on our site.]" ..
			server_status.get_formspec( )

		return formspec
	end

	local function on_close( state, fields )
		if fields.quit == minetest.FORMSPEC_SIGCONT then
			if is_account_created then
				open_connect_viewer( )
			else
				minetest.update_menu( get_formspec( ) )
			end
			return
		elseif fields.quit then
			return  -- ignore other signals
		end

		if fields.register then
			username = fields.username

			if #fields.username < 3 or not string.match( fields.username, "^[a-zA-Z0-9_]+$" ) then
				open_alert_box( "halt", "Error:\n\nPlease enter a valid username.  Only alphanumeric characters and underscores are accepted." )
				return
			elseif #fields.password < 3 then
				open_alert_box( "halt", "Error:\n\nPlease enter a longer password." )
				return
			elseif birthdate.to_age( ) < 16 then
				open_alert_box( "halt", "Error:\n\nYou must be at least 16 years of age to play." )
				return
			end

			local err = register_account( fields.username, fields.password, birthdate.to_birthday( ) )
			if err then
				open_alert_box( "halt", err )
			else
				open_alert_box( "info", "A new account has been created. You can now login with your username and password." )
				is_account_created = true
			end

		elseif fields.menu_text then
			if fields.menu_text == "action:skipstep" then
				open_connect_viewer( )
			elseif fields.menu_text == "action:policy" then
				core.open_url( "http://www.clubequinox.com/policy.html" )
			end

		elseif fields.birthdate_day then
			birthdate.set_day( fields.birthdate_day )

		elseif fields.birthdate_month then
			birthdate.set_month( fields.birthdate_month )

		elseif fields.birthdate_year then
			birthdate.set_year( tonumber( fields.birthdate_year ) )

		end
	end

	core.create_menu( nil, get_formspec( ), on_close )
end

server_status = ServerStatus( )
core.set_clouds( false )

--print( scarlet.translate_72dpi( "image[1,1;195p,40p;" .. fs( defaulttexturedir .. "menu_logo.png" ) .. "]" ) )

open_register_viewer( )

--------------------------------------------------------------------------------
core.event_handler = function(event)
	print( "event", dump( fields ))
end